@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 """ grin searches text files.
 """
+from __future__ import print_function
 
 import bisect
 import fnmatch
@@ -11,12 +12,19 @@ import re
 import shlex
 import stat
 import sys
+from io import UnsupportedOperation
 
 import argparse
 
+if sys.version_info[0] > 2:
+    to_str = lambda s : s.decode('latin1')
+    ints2bytes = bytes
+else:
+    to_str = str
+    ints2bytes = lambda ints : ''.join(map(chr, ints))
 
 #### Constants ####
-__version__ = '1.2'
+__version__ = '1.3.0'
 
 # Maintain the numerical order of these constants. We use them for sorting.
 PRE = -1
@@ -24,8 +32,8 @@ MATCH = 0
 POST = 1
 
 # Use file(1)'s choices for what's text and what's not.
-TEXTCHARS = ''.join(map(chr, [7,8,9,10,12,13,27] + range(0x20, 0x100)))
-ALLBYTES = ''.join(map(chr, range(256)))
+TEXTCHARS = ints2bytes([7,8,9,10,12,13,27] + list(range(0x20, 0x100)))
+ALLBYTES = ints2bytes(range(256))
 
 COLOR_TABLE = ['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan',
                'white', 'default']
@@ -35,12 +43,11 @@ COLOR_STYLE = {
         }
 
 # gzip magic header bytes.
-GZIP_MAGIC = '\037\213'
+GZIP_MAGIC = b'\037\213'
 
 # Target amount of data to read into memory at a time.
 READ_BLOCKSIZE = 16 * 1024 * 1024
 
-
 def is_binary_string(bytes):
     """ Determine if a string is classified as binary rather than text.
 
@@ -54,7 +61,8 @@ def is_binary_string(bytes):
     """
     nontext = bytes.translate(ALLBYTES, TEXTCHARS)
     return bool(nontext)
-    
+
+
 def get_line_offsets(block):
     """ Compute the list of offsets in DataBlock 'block' which correspond to
     the beginnings of new lines.
@@ -80,7 +88,8 @@ def get_line_offsets(block):
             # Keep track of the count of lines within the "current block"
             if next_newline >= block.start and next_newline < block.end:
                 line_count += 1
-            
+
+
 def colorize(s, fg=None, bg=None, bold=False, underline=False, reverse=False):
     """ Wraps a string with ANSI color escape sequences corresponding to the
     style parameters given.
@@ -207,7 +216,7 @@ class GrepText(object):
     def read_block_with_context(self, prev, fp, fp_size):
         """ Read a block of data from the file, along with some surrounding
         context.
-        
+
         Parameters
         ----------
         prev : DataBlock, or None
@@ -216,23 +225,23 @@ class GrepText(object):
 
         fp : filelike object
             The source of block data.
-        
+
         fp_size : int or None
             Size of the file in bytes, or None if the size could not be
             determined.
-        
+
         Returns
         -------
         A DataBlock representing the "current" block along with context.
         """
         if fp_size is None:
             target_io_size = READ_BLOCKSIZE
-            block_main = fp.read(target_io_size)
+            block_main = to_str(fp.read(target_io_size))
             is_last_block = len(block_main) < target_io_size
         else:
             remaining = max(fp_size - fp.tell(), 0)
             target_io_size = min(READ_BLOCKSIZE, remaining)
-            block_main = fp.read(target_io_size)
+            block_main = to_str(fp.read(target_io_size))
             is_last_block = target_io_size == remaining
 
         if prev is None:
@@ -271,12 +280,13 @@ class GrepText(object):
         before_lines = prev.data[before_start:prev.end]
         # Using readline() to force this block out to a newline boundary...
         curr_block = (prev.data[prev.end:] + block_main +
-            ('' if is_last_block else fp.readline()))
+            ('' if is_last_block else to_str(fp.readline())))
         # Read in some lines of 'after' context.
         if is_last_block:
             after_lines = ''
         else:
-            after_lines_list = [fp.readline() for i in range(self.options.after_context)]
+            after_lines_list = [to_str(fp.readline())
+                                for i in range(self.options.after_context)]
             after_lines = ''.join(after_lines_list)
 
         result = DataBlock(
@@ -304,14 +314,19 @@ class GrepText(object):
         """
         context = []
         line_count = 0
-        try:
-            status = os.fstat(fp.fileno())
-            if stat.S_ISREG(status.st_mode):
-                fp_size = status.st_size
-            else:
+        if isinstance(fp, gzip.GzipFile):
+            fp_size = None  # gzipped data is usually longer than the file
+        else:
+            try:
+                file_no = fp.fileno()
+            except (AttributeError, UnsupportedOperation):  # doesn't support fileno()
                 fp_size = None
-        except AttributeError:  # doesn't support fileno()
-            fp_size = None
+            else:
+                status = os.fstat(file_no)
+                if stat.S_ISREG(status.st_mode):
+                    fp_size = status.st_size
+                else:
+                    fp_size = None
 
         block = self.read_block_with_context(None, fp, fp_size)
         while block.end > block.start:
@@ -454,7 +469,7 @@ class GrepText(object):
                         color_substring = colorize(old_substring, **style)
                         line = line[:start] + color_substring + line[end:]
                         total_offset += len(color_substring) - len(old_substring)
-                        
+
                 ns = dict(
                     lineno = i+1,
                     sep = {PRE: '-', POST: '+', MATCH: ':'}[kind],
@@ -492,8 +507,8 @@ class GrepText(object):
             f = sys.stdin
             filename = '<STDIN>'
         else:
-            # 'r' does the right thing for both open ('rt') and gzip.open ('rb')
-            f = opener(filename, 'r')
+            # Always open in binary mode
+            f = opener(filename, 'rb')
         try:
             unique_context = self.do_grep(f)
         finally:
@@ -584,7 +599,7 @@ class FileRecognizer(object):
         """
         try:
             bytes = f.read(self.binary_bytes)
-        except Exception, e:
+        except Exception as e:
             # When trying to read from something that looks like a gzipped file,
             # it may be corrupt. If we do get an error, assume that the file is binary.
             return True
@@ -739,7 +754,7 @@ class FileRecognizer(object):
                 basenames = os.listdir(startpath)
             except OSError:
                 return
-            for basename in basenames:
+            for basename in sorted(basenames):
                 path = os.path.join(startpath, basename)
                 for fn, k in self.walk(path):
                     yield fn, k
@@ -787,7 +802,7 @@ def get_grin_arg_parser(parser=None):
     parser.add_argument('-L', '--files-without-matches', action='store_true',
         dest='show_match', default=False,
         help="show the matches with the filenames")
-    parser.add_argument('--no-color', action='store_true', default=False,
+    parser.add_argument('--no-color', action='store_true', default=sys.platform == 'win32',
         help="do not use colorized output [default if piping the output]")
     parser.add_argument('--use-color', action='store_false', dest='no_color',
         help="use colorized output [default if outputting to a terminal]")
@@ -903,12 +918,15 @@ def get_grind_arg_parser(parser=None):
 def get_recognizer(args):
     """ Get the file recognizer object from the configured options.
     """
+    # Make sure we have empty sets when we have empty strings.
+    skip_dirs = set([x for x in args.skip_dirs.split(',') if x])
+    skip_exts = set([x for x in args.skip_exts.split(',') if x])
     fr = FileRecognizer(
         skip_hidden_files=args.skip_hidden_files,
         skip_backup_files=args.skip_backup_files,
         skip_hidden_dirs=args.skip_hidden_dirs,
-        skip_dirs=set(args.skip_dirs.split(',')),
-        skip_exts=set(args.skip_exts.split(',')),
+        skip_dirs=skip_dirs,
+        skip_exts=skip_exts,
         skip_symlink_files=not args.follow_symlinks,
         skip_symlink_dirs=not args.follow_symlinks,
     )
@@ -1026,9 +1044,15 @@ def grin_main(argv=None):
             sys.stdout.write(report)
     except KeyboardInterrupt:
         raise SystemExit(0)
+    except IOError as e:
+        if 'Broken pipe' in str(e):
+            # The user is probably piping to a pager like less(1) and has exited
+            # it. Just exit.
+            raise SystemExit(0)
+        raise
 
 def print_line(filename):
-    print filename
+    print(filename)
 
 def print_null(filename):
     # Note that the final filename will have a trailing NUL, just like 
@@ -1061,6 +1085,12 @@ def grind_main(argv=None):
                     output(filename)
     except KeyboardInterrupt:
         raise SystemExit(0)
+    except IOError as e:
+        if 'Broken pipe' in str(e):
+            # The user is probably piping to a pager like less(1) and has exited
+            # it. Just exit.
+            raise SystemExit(0)
+        raise
 
 if __name__ == '__main__':
     grin_main()
